package com.yonge.netty.dto;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Optional;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

import javax.sound.sampled.AudioFormat;

import org.apache.commons.lang3.StringUtils;
import org.apache.commons.math3.analysis.function.Gaussian;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.yonge.audio.analysis.Signals;
import com.yonge.audio.analysis.detector.YINPitchDetector;
import com.yonge.netty.dto.NoteAnalysis.NoteErrorType;
import com.yonge.netty.entity.MusicXmlBasicInfo;
import com.yonge.netty.entity.MusicXmlNote;
import com.yonge.netty.entity.MusicXmlSection;
import com.yonge.netty.server.processor.WaveformWriter;

import be.tarsos.dsp.pitch.FastYin;

/**
 * 用户通道上下文
 */
public class UserChannelContext {

	private final static Logger LOGGER = LoggerFactory.getLogger(UserChannelContext.class);

	// 打击乐
	// private final static List<Integer> percussionList = Arrays.asList(23, 113, 121);

	private final static int MIN_FREQUECY = 43;

	private final static int MAX_FREQUECY = 2000;

	private FastYin detector;

	private String user;

	private double standardFrequecy = 442;

	private float offsetMS;

	private float micDelayMS;

	private double dynamicOffset;

	private String platform;

	private Long recordId;

	private Integer subjectId;

	private String evaluationCriteria;

	private float beatDuration;

	private boolean delayProcessed;

	// 曲目与musicxml对应关系
	private ConcurrentHashMap<Integer, MusicXmlBasicInfo> songMusicXmlMap = new ConcurrentHashMap<Integer, MusicXmlBasicInfo>();

	private WaveformWriter waveFileProcessor;

	private NoteAnalysis processingNote = new NoteAnalysis(0, 0, -1);

	private AtomicInteger evaluatingSectionIndex = new AtomicInteger(0);

	private List<NoteAnalysis> doneNoteAnalysisList = new ArrayList<NoteAnalysis>();

	private List<SectionAnalysis> doneSectionAnalysisList = new ArrayList<SectionAnalysis>();

	private List<ChunkAnalysis> totalChunkAnalysisList = new ArrayList<ChunkAnalysis>();

	private byte[] channelBufferBytes = new byte[0];

	private double playTime;

	private HardLevelEnum hardLevel = HardLevelEnum.BEGINNER;

	private boolean handlerSwitch;

	private NotePlayResult queryNoteFrequency(MusicXmlNote xmlNote, double playFrequency) {

		NotePlayResult result = new NotePlayResult();

		boolean status = false;
		double migrationRate = 0;

		if (Math.round(xmlNote.getFrequency()) == Math.round(playFrequency)) {
			status = true;
			migrationRate = 0;
		} else {
			NoteFrequencyRange noteFrequencyRange = new NoteFrequencyRange(standardFrequecy, xmlNote.getFrequency());

			if (noteFrequencyRange.getMinFrequency() > playFrequency || playFrequency > noteFrequencyRange.getMaxFrequency()) {
				status = false;
			} else {

				status = true;

				if (Math.round(playFrequency) < Math.round(xmlNote.getFrequency())) {
					double min = Math.abs(xmlNote.getFrequency() - noteFrequencyRange.getMinFrequency()) / 2;
					migrationRate = Math.abs(playFrequency - xmlNote.getFrequency()) / min;
				} else {
					double max = Math.abs(xmlNote.getFrequency() - noteFrequencyRange.getMaxFrequency()) / 2;
					migrationRate = Math.abs(playFrequency - xmlNote.getFrequency()) / max;
				}
			}
		}

		result.setStatus(status);
		result.setMigrationRate(migrationRate);

		return result;
	}

	public void init(MusicXmlBasicInfo musicXmlBasicInfo, float sampleRate, int bufferSize) {
		this.platform = musicXmlBasicInfo.getPlatform();
		this.subjectId = musicXmlBasicInfo.getSubjectId();
		this.beatDuration = musicXmlBasicInfo.getBeatLength();
		this.hardLevel = HardLevelEnum.valueOf(musicXmlBasicInfo.getHeardLevel());
		this.evaluationCriteria = musicXmlBasicInfo.getEvaluationCriteria();
		if (detector == null) {
			detector = new FastYin(sampleRate, bufferSize);
		}
	}

	public void setUser(String user) {
		this.user = user;
	}

	public Long getRecordId() {
		return recordId;
	}

	public void setRecordId(Long recordId) {
		this.recordId = recordId;
	}

	public boolean getHandlerSwitch() {
		return handlerSwitch;
	}

	public void setHandlerSwitch(boolean handlerSwitch) {
		this.handlerSwitch = handlerSwitch;
	}

	public float getOffsetMS() {
		return offsetMS;
	}

	public void setOffsetMS(float offsetMS) {
		this.offsetMS = offsetMS;
	}

	public float getMicDelayMS() {
		return micDelayMS;
	}

	public void setMicDelayMS(float micDelayMS) {
		this.micDelayMS = micDelayMS;
	}

	public float getBeatDuration() {
		return beatDuration;
	}

	public void setBeatDuration(float beatDuration) {
		this.beatDuration = beatDuration;
	}

	public HardLevelEnum getHardLevel() {
		return hardLevel;
	}

	public ConcurrentHashMap<Integer, MusicXmlBasicInfo> getSongMusicXmlMap() {
		return songMusicXmlMap;
	}

	public WaveformWriter getWaveFileProcessor() {
		return waveFileProcessor;
	}

	public void setWaveFileProcessor(WaveformWriter waveFileProcessor) {
		this.waveFileProcessor = waveFileProcessor;
	}

	public NoteAnalysis getProcessingNote() {
		return processingNote;
	}

	public void setProcessingNote(NoteAnalysis processingNote) {
		this.processingNote = processingNote;
	}

	public List<SectionAnalysis> getDoneSectionAnalysisList() {
		return doneSectionAnalysisList;
	}

	public List<NoteAnalysis> getDoneNoteAnalysisList() {
		return doneNoteAnalysisList;
	}

	public void resetUserInfo() {
		waveFileProcessor = null;
		processingNote = new NoteAnalysis(0, 0, -1);
		evaluatingSectionIndex = new AtomicInteger(0);
		channelBufferBytes = new byte[0];
		doneNoteAnalysisList = new ArrayList<NoteAnalysis>();
		doneSectionAnalysisList = new ArrayList<SectionAnalysis>();
		totalChunkAnalysisList = new ArrayList<ChunkAnalysis>();
		recordId = null;
		playTime = 0;
		delayProcessed = false;
		offsetMS = 0;
		dynamicOffset = 0;
		handlerSwitch = false;
	}

	public MusicXmlBasicInfo getMusicXmlBasicInfo(Integer songId) {
		MusicXmlBasicInfo musicXmlBasicInfo = null;
		if (songId == null) {
			musicXmlBasicInfo = songMusicXmlMap.values().stream().findFirst().get();
		} else {
			musicXmlBasicInfo = songMusicXmlMap.get(songId);
		}
		return musicXmlBasicInfo;
	}

	public MusicXmlSection getCurrentMusicSection(Integer songId, int sectionIndex) {
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);
		return musicXmlBasicInfo.getMusicXmlSectionMap().get(sectionIndex);
	}

	public MusicXmlNote getCurrentMusicNote(Integer songId, Integer noteIndex) {
		if (songMusicXmlMap.size() == 0) {
			return null;
		}
		if (noteIndex == null) {
			noteIndex = processingNote.getMusicalNotesIndex();
		}
		final int index = noteIndex;
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		int totalNoteIndex = getTotalMusicNoteIndex(null);
		if (musicXmlBasicInfo != null && index <= totalNoteIndex) {
			return musicXmlBasicInfo.getMusicXmlInfos().stream().filter(t -> t.getMusicalNotesIndex() == index).findFirst().get();
		}

		return null;
	}

	public int getTotalMusicNoteIndex(Integer songId) {
		if (songMusicXmlMap.size() == 0) {
			return -1;
		}
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		if (musicXmlBasicInfo != null) {
			return musicXmlBasicInfo.getMusicXmlInfos().stream().map(t -> t.getMusicalNotesIndex()).distinct().max(Integer::compareTo).get();
		}

		return -1;
	}

	public List<MusicXmlNote> getCurrentMusicSection(Integer songId, Integer sectionIndex) {
		if (songMusicXmlMap.size() == 0) {
			return null;
		}
		if (sectionIndex == null) {
			sectionIndex = processingNote.getSectionIndex();
		}
		final int index = sectionIndex;
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		if (musicXmlBasicInfo != null) {
			return musicXmlBasicInfo.getMusicXmlInfos().stream().filter(t -> t.getMusicalNotesIndex() == index)
					.sorted(Comparator.comparing(MusicXmlNote::getMusicalNotesIndex)).collect(Collectors.toList());
		}

		return null;
	}

	public int getTotalMusicSectionSize(Integer songId) {
		if (songMusicXmlMap.size() == 0) {
			return -1;
		}
		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		if (musicXmlBasicInfo != null) {
			return (int) musicXmlBasicInfo.getMusicXmlInfos().stream().map(t -> t.getMeasureIndex()).distinct().count();
		}

		return -1;
	}

	public int getMusicSectionIndex(Integer songId, int musicXmlNoteIndex) {
		if (songMusicXmlMap.size() == 0) {
			return -1;
		}

		if (getTotalMusicNoteIndex(null) < musicXmlNoteIndex) {
			return -1;
		}

		MusicXmlBasicInfo musicXmlBasicInfo = getMusicXmlBasicInfo(songId);

		if (musicXmlBasicInfo != null) {
			return musicXmlBasicInfo.getMusicXmlInfos().stream().filter(t -> t.getMusicalNotesIndex() == musicXmlNoteIndex).findFirst().get().getMeasureIndex();
		}

		return -1;
	}

	public byte[] getChannelBufferBytes() {
		return channelBufferBytes;
	}

	public void setChannelBufferBytes(byte[] channelBufferBytes) {
		this.channelBufferBytes = channelBufferBytes;
	}

	public AtomicInteger getEvaluatingSectionIndex() {
		return evaluatingSectionIndex;
	}

	public void handle(float[] samples, AudioFormat audioFormat) {

		// YINPitchDetector frequencyDetector = new YINPitchDetector(samples.length , audioFormat.getSampleRate());
		// int playFrequency = (int) frequencyDetector.getFrequency(samples);

		int playFrequency = -1;
		if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			playFrequency = (int) detector.getPitch(samples).getPitch();
		}

		double splDb = Signals.soundPressureLevel(samples);
		double power = Signals.power(samples);
		int amplitude = (int) Signals.norm(samples);
		
		int decibels = (int) Signals.decibels(samples);
		if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.AMPLITUDE.getCode())) {
			amplitude = (int) Signals.norm(samples);
		} else if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.DECIBELS.getCode())) {
			amplitude = (int) Signals.decibels(samples);
			amplitude = amplitude >= 60 ? amplitude : 0;
		}
		// float rms = Signals.rms(samples);

		double durationTime = 1000 * (samples.length * 2) / audioFormat.getSampleRate() / (audioFormat.getSampleSizeInBits() / 8);

		playTime += durationTime;

		// 获取当前音符信息
		MusicXmlNote musicXmlNote = getCurrentMusicNote(null, null);

		if (musicXmlNote == null) {
			return;
		}

		// 取出当前处理中的音符信息
		NoteAnalysis noteAnalysis = getProcessingNote();
		if (noteAnalysis == null || noteAnalysis.getDurationTime() == 0) {
			noteAnalysis = new NoteAnalysis(musicXmlNote.getMeasureRenderIndex(), musicXmlNote.getMusicalNotesIndex(), musicXmlNote.getMeasureIndex(),
					(int) musicXmlNote.getFrequency(), musicXmlNote.getDuration());
		}

		evaluatingSectionIndex.set(noteAnalysis.getSectionIndex());

		if (noteAnalysis.getMusicalNotesIndex() >= 0 && noteAnalysis.getMusicalNotesIndex() <= getTotalMusicNoteIndex(null)) {

			LOGGER.debug("user:{}  delayProcessed:{}  dynamicOffset:{}  Frequency:{}  splDb:{}  power:{}  amplitude:{} decibels:{}  endtime:{}", user,
					delayProcessed, dynamicOffset, playFrequency, splDb, power, amplitude, decibels, playTime);

			ChunkAnalysis chunkAnalysis = new ChunkAnalysis(playTime - durationTime, playTime, playFrequency, splDb, power, amplitude);

			if (totalChunkAnalysisList.size() > 0) {
				if (totalChunkAnalysisList.get(totalChunkAnalysisList.size() - 1).getAmplitude() + 2 < chunkAnalysis.getAmplitude()) {
					chunkAnalysis.setPeak(true);// 只针对打击乐
				}
			}
			totalChunkAnalysisList.add(chunkAnalysis);

			if (playTime >= (musicXmlNote.getDuration() + musicXmlNote.getTimeStamp())) {

				//musicXmlNote.setTimeStamp(musicXmlNote.getTimeStamp() + micDelayMS);

				if (musicXmlNote.getDontEvaluating()) {
					noteAnalysis.setIgnore(true);
				}
				
				List<ChunkAnalysis> nodeChunkAnalysisList = extract(musicXmlNote, noteAnalysis);

				// 判断节奏(音符持续时间内有不间断的音高，就节奏正确)
				if (!StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
					noteAnalysis.setPlayFrequency(-1);
					noteAnalysis.setTempoStatus(computeTempoWithAmplitude2(musicXmlNote, noteAnalysis, nodeChunkAnalysisList));
				} else {
					noteAnalysis.setPlayFrequency(computeFrequency(musicXmlNote, nodeChunkAnalysisList));
					noteAnalysis.setTempoStatus(computeTempoWithFrequency(musicXmlNote, noteAnalysis, nodeChunkAnalysisList));
				}

				evaluateForNote(musicXmlNote, noteAnalysis, nodeChunkAnalysisList);// 对当前音符评分

				LOGGER.debug("当前音符下标[{}] 预计频率：{} 实际频率：{} 节奏：{}", noteAnalysis.getMusicalNotesIndex(), musicXmlNote.getFrequency(),
						noteAnalysis.getPlayFrequency(), noteAnalysis.getTempoStatus());

				doneNoteAnalysisList.add(noteAnalysis);

				// 准备处理下一个音符
				int nextNoteIndex = musicXmlNote.getMusicalNotesIndex() + 1;
				float nextNoteFrequence = -1;
				double standDuration = 0;
				int measureRenderIndex = 0;
				MusicXmlNote nextMusicXmlNote = getCurrentMusicNote(null, nextNoteIndex);
				if (nextMusicXmlNote != null) {
					nextNoteFrequence = nextMusicXmlNote.getFrequency();
					standDuration = nextMusicXmlNote.getDuration();
					measureRenderIndex = nextMusicXmlNote.getMeasureRenderIndex();
				}

				NoteAnalysis nextNoteAnalysis = new NoteAnalysis(measureRenderIndex, nextNoteIndex, getMusicSectionIndex(null, nextNoteIndex),
						(int) nextNoteFrequence, standDuration);

				noteAnalysis = nextNoteAnalysis;

			}

			setProcessingNote(noteAnalysis);
		}

	}

	public int evaluateForSection(int sectionIndex, int subjectId) {

		int score = -1;
		if (doneSectionAnalysisList.size() >= getTotalMusicSectionSize(null)) {
			return score;
		}

		// 取出当前小节的所有音符
		List<NoteAnalysis> noteAnalysisList = doneNoteAnalysisList.stream().filter(t -> t.getSectionIndex() == sectionIndex).collect(Collectors.toList());

		long ignoreSize = noteAnalysisList.stream().filter(t -> t.isIgnore()).count();

		SectionAnalysis sectionAnalysis = new SectionAnalysis();
		sectionAnalysis.setIndex(sectionIndex);
		sectionAnalysis.setNoteNum(noteAnalysisList.size());
		sectionAnalysis.setIsIngore(ignoreSize == noteAnalysisList.size());

		// 判断是否需要评分
		MusicXmlSection musicXmlSection = getCurrentMusicSection(null, sectionIndex);
		if (noteAnalysisList.size() == musicXmlSection.getNoteNum()) {
			sectionAnalysis.setMeasureRenderIndex(noteAnalysisList.stream().findFirst().get().getMeasureRenderIndex());
			// 取出需要评测的音符
			List<NoteAnalysis> noteList = noteAnalysisList.stream().filter(t -> t.isIgnore() == false).collect(Collectors.toList());

			if (noteList != null && noteList.size() > 0) {
				score = noteList.stream().mapToInt(t -> t.getScore()).sum() / noteList.size();
			}
			sectionAnalysis.setDurationTime(noteAnalysisList.stream().mapToDouble(t -> t.getDurationTime()).sum());
			sectionAnalysis.setScore(score);

			LOGGER.debug("小节评分：{}", sectionAnalysis);
			doneSectionAnalysisList.add(sectionAnalysis);
		}

		return score;
	}

	public Map<String, Integer> evaluateForMusic() {

		Map<String, Integer> result = new HashMap<String, Integer>();

		result.put("playTime", (int) doneNoteAnalysisList.stream().mapToDouble(t -> t.getDurationTime()).sum());
		result.put("recordId", recordId.intValue());

		// 取出需要评测的音符
		List<NoteAnalysis> noteAnalysisList = doneNoteAnalysisList.stream().filter(t -> t.isIgnore() == false).collect(Collectors.toList());

		if (noteAnalysisList != null && noteAnalysisList.size() > 0) {
			int intonationScore = 0;
			int tempoScore = 0;
			int integrityScore = 0;
			int socre = 0;

			for (NoteAnalysis note : noteAnalysisList) {
				intonationScore += note.getIntonationScore();
				tempoScore += note.getTempoScore();
				integrityScore += note.getIntegrityScore();
				socre += note.getScore();
			}

			tempoScore = tempoScore / noteAnalysisList.size();
			intonationScore = intonationScore / noteAnalysisList.size();
			integrityScore = integrityScore / noteAnalysisList.size();

			result.put("cadence", tempoScore);
			result.put("intonation", intonationScore);
			result.put("integrity", integrityScore);

			int score = socre / noteAnalysisList.size();

			// 平均得分
			if (getMusicXmlBasicInfo(null).getSubjectId() == 23 || getMusicXmlBasicInfo(null).getSubjectId() == 113) {
				score = tempoScore;
			}
			result.put("score", score);
		} else {
			result.put("cadence", 0);
			result.put("intonation", 0);
			result.put("integrity", 0);
			result.put("score", 0);
		}
		return result;
	}

	public void evaluateForNote(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> chunkAnalysisList) {

		double durationTime = chunkAnalysisList.get(chunkAnalysisList.size() - 1).getEndTime() - chunkAnalysisList.get(0).getStartTime();

		double playDurationTime = 0;

		if (!StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			if (noteAnalysis.getFrequency() == -1) {// 休止符
				if (noteAnalysis.getTempoStatus() == 0) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
				} else {
					noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
				}
			} else {
				int beatTimes = (int) chunkAnalysisList.stream().filter(t -> t.getAmplitude() > 0).count();
				if (beatTimes == 0) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.NOT_PLAY);
				} else if (noteAnalysis.getTempoStatus() == 0) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
				} else {
					noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
				}
			}
		} else {

			NotePlayResult notePlayResult = queryNoteFrequency(musicXmlNote, noteAnalysis.getPlayFrequency());

			if (noteAnalysis.getFrequency() == -1) {// 休止符

				playDurationTime = chunkAnalysisList.stream().filter(t -> t.getFrequency() <= MIN_FREQUECY).mapToDouble(t -> t.getDurationTime()).sum();

				if (noteAnalysis.getTempoStatus() == 0) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
				} else if (playDurationTime * 100 / durationTime < hardLevel.getIntegrityRange()) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.INTEGRITY_WRONG);
				} else if (notePlayResult.getStatus() == false) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.INTONATION_WRONG);
				} else {
					noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
				}
			} else {
				playDurationTime = chunkAnalysisList.stream().filter(t -> t.getFrequency() > MIN_FREQUECY && t.getFrequency() < MAX_FREQUECY)
						.mapToDouble(t -> t.getDurationTime()).sum();

				if (playDurationTime * 100 / durationTime < hardLevel.getNotPlayRange()) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.NOT_PLAY);
					LOGGER.debug("未演奏：{}", playDurationTime * 100 / durationTime);
				} else if (playDurationTime * 100 / durationTime < hardLevel.getIntegrityRange()) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.INTEGRITY_WRONG);
					LOGGER.debug("完整度不足：{}", playDurationTime * 100 / durationTime);
				} else if (noteAnalysis.getTempoStatus() == 0) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.CADENCE_WRONG);
				} else if (notePlayResult.getStatus() == false) {
					noteAnalysis.setMusicalErrorType(NoteErrorType.INTONATION_WRONG);
				} else {
					noteAnalysis.setMusicalErrorType(NoteErrorType.RIGHT);
				}
			}

		}

		// 计算音分
		int tempoScore = 0;
		int integrityScore = 0;
		int intonationScore = 100 - new BigDecimal(Math
				.abs(YINPitchDetector.hertzToAbsoluteCent(noteAnalysis.getPlayFrequency()) - YINPitchDetector.hertzToAbsoluteCent(noteAnalysis.getFrequency())))
						.multiply(new BigDecimal(20)).divide(new BigDecimal(17), BigDecimal.ROUND_UP).setScale(0, BigDecimal.ROUND_UP).intValue();
		if (intonationScore < 0) {
			intonationScore = 0;
		} else if (intonationScore > 100) {
			intonationScore = 100;
		}

		if (noteAnalysis.getMusicalErrorType() == NoteErrorType.NOT_PLAY) {
			intonationScore = 0;
		} else {

			if (noteAnalysis.getTempoStatus() != 0) {
				tempoScore = 100;
				noteAnalysis.setTempoScore(tempoScore);
			}

			integrityScore = (int) (playDurationTime * 100 * 100 / hardLevel.getIntegrityRange() / durationTime);
			if (integrityScore > 100) {
				integrityScore = 100;
			}
			noteAnalysis.setIntegrityScore(integrityScore);
		}
		noteAnalysis.setIntonationScore(intonationScore);
		if (!StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			noteAnalysis.setScore(tempoScore);
		} else {
			noteAnalysis.setScore(
					new BigDecimal(intonationScore + tempoScore + integrityScore).divide(new BigDecimal(3), 2).setScale(0, BigDecimal.ROUND_UP).intValue());
		}
	}

	private int computeFrequency(MusicXmlNote musicXmlNote, List<ChunkAnalysis> chunkAnalysisList) {

		if (chunkAnalysisList == null || chunkAnalysisList.size() == 0) {
			return -1;
		}

		reduceNoise(chunkAnalysisList, EvaluationCriteriaEnum.FREQUENCY);

		List<ChunkAnalysis> chunkList = new ArrayList<ChunkAnalysis>(chunkAnalysisList);

		List<Integer> chunkFrequencyList = chunkList.stream().map(t -> t.getFrequency())
				.filter(t -> t.doubleValue() > MIN_FREQUECY && t.doubleValue() < MAX_FREQUECY).collect(Collectors.toList());

		if (chunkFrequencyList.size() == 0) {
			return -1;
		}

		int frequency = (int) (chunkFrequencyList.stream().mapToInt(t -> t).sum() / chunkFrequencyList.size());
		
		return frequency;
	}

	/**
	 * 时值范围内有且只有一个音，且不能间断，且在合理范围内需开始演奏
	 * 与上一个音相同时，2个音之间需要间断
	 * @param musicXmlNote
	 * @return
	 */
	private int computeTempoWithFrequency(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> chunkAnalysisList) {
		
		LOGGER.debug("---------------------TEMPO----------------------");
		
		double avgPower = chunkAnalysisList.stream().collect(Collectors.averagingDouble(ChunkAnalysis::getSplDb));
		noteAnalysis.setDBSPL(avgPower);

		if (noteAnalysis.getTempoStatus() == 0 || chunkAnalysisList == null || chunkAnalysisList.size() == 0) {
			return 0;
		}

		reduceNoise(chunkAnalysisList, EvaluationCriteriaEnum.FREQUENCY);

		if (musicXmlNote.getFrequency() == -1) {// 休止符
			return chunkAnalysisList.stream().filter(t -> t.getFrequency() > MIN_FREQUECY).count() <= 1 ? 1 : 0;
		}

		// 将信号分堆归类
		Map<Integer, Integer> signalGrouping = new HashMap<Integer, Integer>();
		for (int i = 1; i < chunkAnalysisList.size(); i++) {
			if (isSamePitch(chunkAnalysisList.get(i).getFrequency(), chunkAnalysisList.get(i - 1).getFrequency())) {
				if (signalGrouping.get(chunkAnalysisList.get(i - 1).getFrequency()) == null) {
					signalGrouping.put(chunkAnalysisList.get(i - 1).getFrequency(), 1);
				} else {
					signalGrouping.put(chunkAnalysisList.get(i - 1).getFrequency(), signalGrouping.get(chunkAnalysisList.get(i - 1).getFrequency()) + 1);
				}
			} else {
				signalGrouping.put(chunkAnalysisList.get(i - 1).getFrequency(), 1);
			}
		}

		int maxTimes = 0, avgFrequency = 0;

		for (Entry<Integer, Integer> entry : signalGrouping.entrySet()) {
			if (entry.getValue() > maxTimes) {
				maxTimes = entry.getValue();
				avgFrequency = entry.getKey();
			}
		}

		LOGGER.debug("当前音符时值范围内平均音高[{}] 声压[{}]", avgFrequency, avgPower);

		int firstBeatIndex = -1;
		double firstBeatTime = 0;

		int depthThreshold = 2;
		int peakNum = 0;
		int continueNums = 0;
		boolean isContinue = false;
		// 检测音高是否间断
		for (int i = 0; i < chunkAnalysisList.size(); i++) {
			if (firstBeatIndex == -1) {
				if (chunkAnalysisList.get(i).getFrequency() > MIN_FREQUECY || musicXmlNote.getFrequency() == -1) {
					firstBeatIndex = i;
					firstBeatTime = chunkAnalysisList.get(i).getStartTime();
				}
			}
			if (!isSamePitch(avgFrequency, chunkAnalysisList.get(i).getFrequency())) {
				++continueNums;
				if (continueNums >= depthThreshold) {
					if (isContinue == false) {
						LOGGER.debug("范围内查询到异常音高信号,StartTime:{}  CurrentFrequency:{} AvgFrequency:{}", chunkAnalysisList.get(i).getStartTime(), chunkAnalysisList.get(i).getFrequency(), avgFrequency);
						peakNum++;
						isContinue = true;
					}
				}
			} else {
				continueNums = 0;
				isContinue = false;
			}
		}

		if (peakNum > 0) {
			LOGGER.debug("已找到节奏起始点，根据音高又检测到[{}]个断点", peakNum);
			return 0;
		}
		
		// 判断进入时间点
		if (firstBeatIndex != -1 && firstBeatIndex * 100 / (chunkAnalysisList.size() - 1) > hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(), musicXmlNote.getDuration())) {
			LOGGER.debug("节奏(音高)错误原因：进入时间点[{}]太晚", firstBeatTime);
			return 0;
		}
		
		//判断是否与上一个音延续下来的

		ChunkAnalysis firstChunkAnalysis = chunkAnalysisList.get(0);

		Optional<ChunkAnalysis> chunkAnalysisOptional = totalChunkAnalysisList.stream()
				.filter(t -> Double.doubleToLongBits(t.getEndTime()) <= Double.doubleToLongBits(firstChunkAnalysis.getStartTime()))
				.reduce((first, second) -> second);

		ChunkAnalysis lastChunkAnalysis = null;
		if (chunkAnalysisOptional.isPresent()) {
			lastChunkAnalysis = chunkAnalysisOptional.get();
		}

		if (lastChunkAnalysis == null) {
			lastChunkAnalysis = new ChunkAnalysis(0, 0, -1, -40, 0, 0);
		}
		
		//只有音高节奏点才需要检测这个条件
		if(noteAnalysis.getTempoStatus() == 1) {
			if(firstChunkAnalysis.getFrequency() > MIN_FREQUECY && lastChunkAnalysis.getFrequency() > MIN_FREQUECY){
				if(isSamePitch(firstChunkAnalysis.getFrequency(), lastChunkAnalysis.getFrequency())){
					LOGGER.debug("节奏错误原因：上一个音[{}]延续下来导致的", lastChunkAnalysis.getFrequency());
					return 0;
				}
			}
		}
		
		continueNums = 0;
		firstBeatIndex = -1;
		firstBeatTime = 0;
		// 判断过程中声音是否有起伏
		
		//获取上一个音符的声压等级
		Optional<NoteAnalysis> preNoteAnalysisOptinal = doneNoteAnalysisList.stream().filter(t -> t.getIndex() == musicXmlNote.getMusicalNotesIndex() - 1).findFirst();
		if(preNoteAnalysisOptinal.isPresent()) {
			avgPower = preNoteAnalysisOptinal.get().getDBSPL();
		}
		
		chunkAnalysisList.add(0, lastChunkAnalysis);
		Map<Integer, ChunkAnalysis> peakMap = detectPeaks(chunkAnalysisList, avgPower + 2);
		peakNum = peakMap.size();
		if (peakNum > 0) {
			firstBeatIndex = peakMap.keySet().stream().min(Integer::compare).get();
			firstBeatTime = peakMap.get(firstBeatIndex).getStartTime();
		}

		if (peakNum > 1) {
			LOGGER.debug("根据声压检测到[{}]个断点", peakNum);
			return 0;
		}

		// 判断进入时间点
		if (firstBeatIndex * 100 / (chunkAnalysisList.size() - 1) > hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(), musicXmlNote.getDuration())) {
			LOGGER.debug("节奏(声压)错误原因：进入时间点[{}]太晚", firstBeatTime);
			return 0;
		}

		return 99;
	}

	private int computeTempoWithAmplitude2(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis, List<ChunkAnalysis> chunkAnalysisList) {

		if (chunkAnalysisList == null || chunkAnalysisList.size() == 0) {
			return 0;
		}
		//计算平均振幅
		double avgAmplitude = chunkAnalysisList.stream().collect(Collectors.averagingDouble(ChunkAnalysis::getAmplitude));
		
		noteAnalysis.setAmplitude(avgAmplitude);
		noteAnalysis.setDBSPL(chunkAnalysisList.stream().collect(Collectors.averagingDouble(ChunkAnalysis::getSplDb)));

		reduceNoise(chunkAnalysisList, EvaluationCriteriaEnum.AMPLITUDE);
		
		if (musicXmlNote.getFrequency() == -1) {// 休止符

			LOGGER.debug("--Amplitude:{}  Denominator:{}", chunkAnalysisList.stream().map(t -> t.getAmplitude()).collect(Collectors.toList()),
					musicXmlNote.getDenominator());
			return chunkAnalysisList.stream().filter(t -> t.getAmplitude() > 0).count() <= 0 ? 3 : 0;
		}

		ChunkAnalysis firstChunkAnalysis = chunkAnalysisList.get(0);

		Optional<ChunkAnalysis> chunkAnalysisOptional = totalChunkAnalysisList.stream()
				.filter(t -> Double.doubleToLongBits(t.getEndTime()) <= Double.doubleToLongBits(firstChunkAnalysis.getStartTime()))
				.reduce((first, second) -> second);

		ChunkAnalysis lastChunkAnalysis = new ChunkAnalysis(0, 0, -1, -40, 0, 0);
		
		if (chunkAnalysisOptional.isPresent()) {
			lastChunkAnalysis = chunkAnalysisOptional.get();
		}

		chunkAnalysisList.add(0, lastChunkAnalysis);

		// 检测是否有多个波峰
		boolean tempo = false;
		int peakNum = 0;
		int firstBeatIndex = -1;
		double firstBeatTime = 0;

		/**
		boolean isContinue = false;
		int firstPeakIndex = -1;
		int firstPeakValue = 0;
		// int range = hardLevel.getAmplitudeThreshold();
		int range = 5;

		if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.DECIBELS.getCode())) {
			range = 50;
		}

		for (int i = 1; i < chunkAmplitudeList.size(); i++) {
			if (chunkAmplitudeList.get(i - 1) + range >= chunkAmplitudeList.get(i)) {
				isContinue = false;
				continue;
			}

			if (isContinue == false && chunkAmplitudeList.get(i - 1) + range < chunkAmplitudeList.get(i)) {
				isContinue = true;
				peakSize++;

				if (firstPeakIndex == -1) {
					firstPeakIndex = i;
					firstPeakValue = chunkAmplitudeList.get(i);
				}
			}
		}*/
		
		//获取上一个音符的振幅
		Optional<NoteAnalysis> preNoteAnalysisOptinal = doneNoteAnalysisList.stream().filter(t -> t.getIndex() == musicXmlNote.getMusicalNotesIndex() - 1).findFirst();
		if(preNoteAnalysisOptinal.isPresent()) {
			avgAmplitude = preNoteAnalysisOptinal.get().getDBSPL();
		}
		
		Map<Integer, ChunkAnalysis> peakMap = detectPeaks(chunkAnalysisList, avgAmplitude);
		peakNum = peakMap.size();
		if (peakNum > 0) {
			firstBeatIndex = peakMap.keySet().stream().min(Integer::compare).get();
			firstBeatTime = peakMap.get(firstBeatIndex).getStartTime();
		}

		if (peakNum == 0) {
			tempo = lastChunkAnalysis.isPeak();
		} else if (peakNum == 1) {
			tempo = true;
		} else {
			tempo = false;
			LOGGER.debug("有多个波峰");
		}

		if (tempo) {
			// 判断进入时间点
			if (firstBeatIndex * 100 / (chunkAnalysisList.size() - 1) > hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(),
					musicXmlNote.getDuration())) {
				LOGGER.debug("节奏错误原因：进入时间点[{}]太晚", firstBeatTime);
				tempo = false;
			}
		}
		

		return tempo == false ? 0 : 3;
	}

	private double queryFirstNoteStartTime(List<ChunkAnalysis> chunkAnalysisList, MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis) {

		if (chunkAnalysisList == null || chunkAnalysisList.size() == 0) {
			LOGGER.debug("找不到数据,StartTime:{}", musicXmlNote.getTimeStamp() + dynamicOffset);
			return musicXmlNote.getTimeStamp() + dynamicOffset;
		}

		double onsetStartTime = 0;
		double preNoteAvgPower = 0;
		
		Map<Integer, ChunkAnalysis> peakMap = new HashMap<Integer, ChunkAnalysis>();
		
		/**
		ChunkAnalysis firstChunkAnalysis = chunkAnalysisList.get(0);
		
		Optional<ChunkAnalysis> chunkAnalysisOptional = totalChunkAnalysisList.stream()
				.filter(t -> Double.doubleToLongBits(t.getEndTime()) <= Double.doubleToLongBits(firstChunkAnalysis.getStartTime()))
				.reduce((first, second) -> second);

		ChunkAnalysis lastChunkAnalysis = null;
		if (chunkAnalysisOptional.isPresent()) {
			lastChunkAnalysis = chunkAnalysisOptional.get();
		}

		if (lastChunkAnalysis == null) {
			lastChunkAnalysis = new ChunkAnalysis(0, 0, -1, -40, 0, 0);
		}
		
		chunkAnalysisList.add(0, lastChunkAnalysis);
		**/
		
		//获取上一个音符
		Optional<NoteAnalysis> preNoteAnalysisOptinal = doneNoteAnalysisList.stream().filter(t -> t.getIndex() == musicXmlNote.getMusicalNotesIndex() - 1).findFirst();
		
		if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			//有音高，声压才有效
			double preNoteAvgFrequency = -1;
			
			if(preNoteAnalysisOptinal.isPresent()) {
				preNoteAvgPower = preNoteAnalysisOptinal.get().getDBSPL();
				preNoteAvgFrequency = preNoteAnalysisOptinal.get().getPlayFrequency();
			} else {
				preNoteAvgPower = -40;
			}
			
			LOGGER.debug("上一个音符的平均音高[{}] 声压[{}]", preNoteAvgFrequency, preNoteAvgPower);
			
			ChunkAnalysis chunkAnalysis = null;
			
			//既要找到起始点，又要判断节奏是否正确
			//若是休止符
			if(musicXmlNote.getFrequency() < MIN_FREQUECY) {
				for(int i = 0; i < chunkAnalysisList.size(); i++) {
					chunkAnalysis = chunkAnalysisList.get(i);
					
					if(chunkAnalysis.getFrequency() < MIN_FREQUECY) {
						onsetStartTime = chunkAnalysis.getStartTime();
						noteAnalysis.setTempoStatus(1);
						LOGGER.debug("范围内查询到音高信号,EndTime:{} ChunkIndex:{}", onsetStartTime, i);
						return onsetStartTime;
					}
				}
			}
			
			//若不是休止符
			if(musicXmlNote.getFrequency() >= MIN_FREQUECY) {
				for(int i = 0; i < chunkAnalysisList.size(); i++) {
					chunkAnalysis = chunkAnalysisList.get(i);
					
					if(chunkAnalysis.getFrequency() >= MIN_FREQUECY) {
						if (preNoteAvgFrequency < MIN_FREQUECY || !isSamePitch(preNoteAvgFrequency, chunkAnalysis.getFrequency())) {
							onsetStartTime = chunkAnalysis.getStartTime();
							noteAnalysis.setTempoStatus(1);
							LOGGER.debug("范围内查询到音高信号,EndTime:{} ChunkIndex:{}", onsetStartTime, i);
							return onsetStartTime;
						}
					}
				}
			}
			
			peakMap = detectPeaks(chunkAnalysisList, preNoteAvgPower + 2);
			
		} else {

			if(preNoteAnalysisOptinal.isPresent()) {
				preNoteAvgPower = preNoteAnalysisOptinal.get().getDBSPL();
			} else {
				preNoteAvgPower = -40;
			}
			
			peakMap = detectPeaks(chunkAnalysisList, preNoteAvgPower);
		}
		
		if (peakMap.size() > 0) {
			Integer firstPeakIndex = peakMap.keySet().stream().max(Integer::compare).get();
			onsetStartTime = chunkAnalysisList.get(firstPeakIndex).getStartTime();
			LOGGER.debug("范围内查询到声压信号,StartTime:{}", onsetStartTime);
			noteAnalysis.setTempoStatus(2);
			return onsetStartTime;
		}

		if (onsetStartTime == 0) {
			LOGGER.debug("范围内未查询到信号,StartTime:{}", musicXmlNote.getTimeStamp() + dynamicOffset);
		}

		// return chunkAnalysisList.get(chunkAnalysisList.size() - 1).getEndTime();
		return Math.max(musicXmlNote.getTimeStamp() + dynamicOffset, onsetStartTime);
	}

	
	private void reduceNoise(List<ChunkAnalysis> chunkAnalysisList, EvaluationCriteriaEnum criteria) {

		ChunkAnalysis chunkAnalysis = null;

		for (int i = 1; i < chunkAnalysisList.size(); i++) {
			if (i < chunkAnalysisList.size() - 1) {
				chunkAnalysis = chunkAnalysisList.get(i);

				if (EvaluationCriteriaEnum.AMPLITUDE == criteria) {
					if (chunkAnalysisList.get(i - 1).getAmplitude() == 0 && chunkAnalysisList.get(i + 1).getAmplitude() == 0
							&& chunkAnalysis.getAmplitude() > 0) {

						chunkAnalysis.setAmplitude(0);
						// chunkAnalysisList.set(i, chunkAnalysis);
					}
				} else if (EvaluationCriteriaEnum.FREQUENCY == criteria) {
					if (chunkAnalysisList.get(i - 1).getFrequency() == -1 && chunkAnalysisList.get(i + 1).getFrequency() == -1
							&& chunkAnalysis.getFrequency() > 0) {

						chunkAnalysis.setFrequency(-1);
						// chunkAnalysisList.set(i, chunkAnalysis);
					}
				}
			}
		}

	}

	private boolean isSamePitch(double frequency1, double frequency2) {
		if (frequency1 == frequency2) {
			return true;
		}
		return (frequency1 != -1 || frequency2 != -1) && Math.abs(YINPitchDetector.getDeviationCent(frequency1, frequency2)) < 50;
	}

	private Map<Integer, ChunkAnalysis> detectPeaks(List<ChunkAnalysis> signalList, double threshold) {
		Map<Integer, ChunkAnalysis> peaks = new HashMap<Integer, ChunkAnalysis>();
		
		movingAverageSmooth(signalList, 10);
		
		/**
		int waveCrestNums = 0,waveTroughNums = 0;
		for (int i = 1; i < signalList.size() - 1; i++) {
			if (signalList.get(i).getSplDb() > signalList.get(i - 1).getSplDb() && signalList.get(i).getSplDb() >= signalList.get(i + 1).getSplDb()
					&& signalList.get(i).getSplDb() > threshold + 2) {
				waveCrestNums++;
				waveTroughNums = 0;
				if (waveCrestNums > 1) {
					peaks.put(i, signalList.get(i));
					LOGGER.debug("范围内查询到异常音波信号,StartTime:{} CurrentSplDb:{} Threshold:{}", signalList.get(i).getStartTime(), signalList.get(i).getSplDb(), threshold);
				}
			} else if (signalList.get(i).getSplDb() < signalList.get(i - 1).getSplDb() && signalList.get(i).getSplDb() <= signalList.get(i + 1).getSplDb()
					&& signalList.get(i).getSplDb() + 2 < threshold) {
				waveTroughNums++;
				waveCrestNums = 0;
				if (waveTroughNums >= 1) {
					peaks.put(i, signalList.get(i));
					LOGGER.debug("范围内查询到异常音波信号,StartTime:{} CurrentSplDb:{} Threshold:{}", signalList.get(i).getStartTime(), signalList.get(i).getSplDb(), threshold);
				}
			} else {
				waveCrestNums = 0;
				waveTroughNums = 0;
			}
		}*/
		
		int continueSignalThreshold = 2;
		int continueNums = 0;
		int disconnectNums = 0;
		boolean isContinue = false;

		if (StringUtils.equalsIgnoreCase(evaluationCriteria, EvaluationCriteriaEnum.FREQUENCY.getCode())) {
			for (int i = 0; i < signalList.size(); i++) {
				//如果连续2个超过平均值
				if(Math.abs(threshold - signalList.get(i).getSplDb()) > 0) {
					++continueNums;
					if(continueNums >= continueSignalThreshold) {
						disconnectNums = 0;
						if(continueNums != i+1) {
							if(isContinue == false) {
								peaks.put(i, signalList.get(i));
								LOGGER.debug("【过程】范围内查询到声压信号,StartTime:{} CurrentSplDb:{} Threshold:{}" , signalList.get(i).getStartTime(), signalList.get(i).getSplDb(), threshold);
								isContinue = true;
							}
						}
					}
				} else {
					disconnectNums++;
					if(disconnectNums >= continueSignalThreshold) {
						continueNums = 0;
						isContinue = false;
					}
				}
			}
		} else {
			int waveCrestNums = 0;
			for (int i = 1; i < signalList.size() - 1; i++) {
				if (signalList.get(i).getSplDb() > signalList.get(i - 1).getSplDb() && signalList.get(i).getSplDb() >= signalList.get(i + 1).getSplDb()
						&& signalList.get(i).getSplDb() > threshold) {
					waveCrestNums++;
					if (waveCrestNums > 0) {
						peaks.put(i, signalList.get(i));
						LOGGER.debug("范围内查询到异常音波信号,StartTime:{} CurrentSplDb:{} Threshold:{}", signalList.get(i).getStartTime(), signalList.get(i).getSplDb(), threshold);
					}
				} else {
					waveCrestNums = 0;
				}
			}
		}
		
		return peaks;
	}
	
	private List<ChunkAnalysis> extract(MusicXmlNote musicXmlNote, NoteAnalysis noteAnalysis) {
		
		LOGGER.debug("---------------------Extract Data----------------------");

		double floatingRange = musicXmlNote.getDuration() * hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(), musicXmlNote.getDuration()) / 100;

		double endTime = musicXmlNote.getTimeStamp() + dynamicOffset + floatingRange;
		double startTime = musicXmlNote.getTimeStamp() + dynamicOffset - floatingRange;
		
		LOGGER.debug("在范围内寻找起始点——floatingRange:{}  modified [ {} - {} ]", floatingRange, startTime, endTime);

		List<ChunkAnalysis> chunkAnalysisList = totalChunkAnalysisList.stream()
				.filter(t -> Double.doubleToLongBits(t.getStartTime()) >= Double.doubleToLongBits(startTime)
						&& Double.doubleToLongBits(t.getEndTime()) <= Double.doubleToLongBits(endTime))
				.collect(Collectors.toList());

		/**
		 * 过程中找到了起始点都可以作为节奏点，后面再出现波峰波谷就表明节奏错误；否则，后面需要出现一个波峰或波谷
		 * 是否要加上前一个音符的最后一个信号？？？
		 */
		double correctedStartTime = queryFirstNoteStartTime(chunkAnalysisList, musicXmlNote, noteAnalysis);
		double correctedEndTime = correctedStartTime + musicXmlNote.getDuration();
		
		LOGGER.debug("在范围内寻找到起始点，调整后的信号时值范围[ {} - {} ]", correctedStartTime, correctedEndTime);

		if (correctedStartTime != startTime) {
			if (startTime < 0) {
				dynamicOffset = correctedStartTime;
			} else {
				dynamicOffset = correctedStartTime - startTime;
			}
		}

		chunkAnalysisList = totalChunkAnalysisList.stream().filter(t -> Double.doubleToLongBits(t.getStartTime()) >= Double.doubleToLongBits(correctedStartTime)
				&& Double.doubleToLongBits(t.getEndTime()) <= Double.doubleToLongBits(correctedEndTime)).collect(Collectors.toList());

		// 根据完整度取部分有效信号
		int elementSize = chunkAnalysisList.size() * (100 - hardLevel.getTempoEffectiveRange(musicXmlNote.getDenominator(), musicXmlNote.getDuration())) / 100;
		
		List<ChunkAnalysis> datas = chunkAnalysisList.subList(0, elementSize);
		
		LOGGER.debug("根据用户的评测难度，当前音符有效的信号时值范围[ {} - {} ]", chunkAnalysisList.get(0).getStartTime(), chunkAnalysisList.get(elementSize - 1).getEndTime());
		
		return datas;
	}
	
    public void gaussianSmooth(List<ChunkAnalysis> signalList, int windowSize) {
        double sigma = windowSize / 3.0;
        Gaussian gaussian = new Gaussian(1.0, windowSize / 2.0, sigma);
        double[] kernel = new double[windowSize];
        for (int i = 0; i < windowSize; i++) {
            kernel[i] = gaussian.value(i);
        }

        for (int i = 0; i < signalList.size(); i++) {
            double sum = 0;
            double weightSum = 0;
            for (int j = 0; j < windowSize; j++) {
                int index = i + j - windowSize / 2;
                if (index >= 0 && index < signalList.size()) {
                    sum += signalList.get(index).getSplDb() * kernel[j];
                    weightSum += kernel[j];
                }
            }
            signalList.get(i).setSplDb(sum / weightSum);
        }
    }
    
    public void movingAverageSmooth(List<ChunkAnalysis> signalList, int windowSize) {
        for (int i = 0; i < signalList.size(); i++) {
            int start = Math.max(0, i - windowSize / 2);
            int end = Math.min(signalList.size(), i + windowSize / 2);
            double sum = 0;
            for (int j = start; j < end; j++) {
                sum += signalList.get(j).getSplDb();
            }
            signalList.get(i).setSplDb(sum / (end - start));
        }
    }
}
